/* Copyright (c) 2011, FoxTab. (team@foxtab.com)
 * All rights reserved.
 * See License.txt for further details.
 */
const EXTENSION_ID = '{ef4e370e-d9f0-4e00-b93e-a4f274cfdd5a}'
/**
 * Import this module for creating loggers
 * Components.utils.import("resource://[modulespath]/LoggerFactory.jsm");
 * @author Eyal
 */
const EXPORTED_SYMBOLS = ["getLogger"]
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;

// public functions
function getLogger(logId,logPath,fileName,numOfHistoryLogs,dumpEnvironment){
	if ($loggers[logId])
		return $loggers[logId];
	var logger = $createLogger();
	logger.init(logPath,fileName,numOfHistoryLogs);
	$loggers[logId] = logger;
	if (dumpEnvironment)
		$dumpEnvironment(logger);
	return logger;
}

// Private static variables
const $loggers={};
const DirService = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties);
const LEVEL_TXT = ["ERROR   ","WARNING ","INFO    ","DEBUG   "];
const $ERROR = 0;
const $WARNING = 1;
const $INFO = 2;
const $DEBUG = 3;

// Private static functions
function $dumpEnvironment(logger){
	if (logger.disabled)
		return;
	var Application = Cc["@mozilla.org/fuel/application;1"].getService(Ci.fuelIApplication);
	var appPrefs = Application.prefs;
	var extensions = Application.extensions.all;
	var extension = Application.extensions.get(EXTENSION_ID);
	var extPrefs = extension.prefs.all;
	var appInfo = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo);
	var runtime = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULRuntime);
	var ProfD = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
	var TmpD = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("TmpD", Ci.nsIFile);
	// remember logger settings
	var _level = logger.level;
	var _showLevel = logger.showLevel;
	var _showTime = logger.showTime;
	// override logger settings
	logger.level=logger.INFO;
	logger.showLevel = false;
	logger.showTime = false;
	// dump environment
	logger.info("//================================================");
	// extension
	logger.info("name\t\t: "+extension.name);
	logger.info("id\t\t\t: "+extension.id);
	logger.info("version\t\t: "+extension.version);
	logger.info("firstRun\t: "+extension.firstRun);	
	// System
	logger.info("\n//platform");
	logger.info("name             \t: "+appInfo.name);
	logger.info("type             \t: "+appInfo.ID);
	logger.info("version          \t: "+appInfo.version);
	logger.info("vendor           \t: "+appInfo.vendor);
	logger.info("build            \t: "+appInfo.appBuildID);
	logger.info("platform         \t: "+appInfo.platformVersion);
	logger.info("platform build   \t: "+appInfo.platformBuildID);
	logger.info("profile directory\t: "+ProfD.path);
	logger.info("temp directory   \t: "+ProfD.path);
	logger.info("operating system \t: "+runtime.OS);
	logger.info("safe mode        \t: "+runtime.inSafeMode);
	// dump major prefs
	logger.info("\n//system preferences");
	logger.info("general.useragent.locale\t: "+appPrefs.getValue("general.useragent.locale",null));
	logger.info("intl.accept_languages\t\t: "+appPrefs.getValue("intl.accept_languages",null));
	// 
	logger.info("\n//extension modified prefs");
	for (var i=0,length=extPrefs.length;i<length;i++){
		var pref=extPrefs[i];
		if (pref.modified)
			logger.info(pref.name+" "+pref.value);
	}
	logger.info("\n//extension default prefs");
	for (var i=0,length=extPrefs.length;i<length;i++){
		var pref=extPrefs[i];
		if (!pref.modified)
			logger.info(pref.name+" "+pref.value);
	}	
	// extensions
	logger.info("\n//enabled extensions");
	for (var i=0,length=extensions.length;i<length;i++){
		var ext=extensions[i];
		if (ext.enabled)
			logger.info(ext.name+" "+ext.version);
	}
	logger.info("\n//disabled extensions");
	for (var i=0,length=extensions.length;i<length;i++){
		var ext=extensions[i];
		if (!ext.enabled)
			logger.info(ext.name+" "+ext.version);
	}
	logger.info("//================================================");
	// restore logger settings
	logger.level=_level;
	logger.showLevel = _showLevel;
	logger.showTime = _showTime;
}

function $createLogger(){
	// Private variables	
	var $pathToLogFolder = [];
	var $registered = false;
	var $parentLogger = null;
	var $profile = null
	var $numOfHistoryLogs = 0;
	var $converter = null;
	var $file = null;
	
	// Private functions
	function getLogFolder(){
		var file = DirService.get("ProfD", Ci.nsIFile);
		for (var i=0,length=$pathToLogFolder.length;i<length;i++)
			file.append($pathToLogFolder[i]);
		return file;
	}

	/**
	 * Push a log (logname.log.1 => logname.log.2)
	 * @recursive
	 */
	function pushLog(index,last){
		var baseLogName=$profile;
		var file = getLogFolder();
		if (index < 0)
			return;
		if (index > 0)
			file.append(baseLogName+"."+index);
		else
			file.append(baseLogName);
		/*
		if (file.isDirectory())
			; // error
		
			file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
		*/
		if (file.exists()){
			if (last)	 // stale log, delete it
				file.remove(true);
			else	// advance log
				file.moveTo(null,baseLogName+"."+(index+1));
		}
		if (index > 0)
			pushLog(index-1)	// advance previous log
	}

	return {
		// variables
		level		: $WARNING,
		showLevel	: true,
		showDate	: true,
		disabled	: false,
		// constants
		// Every level also log the less severe leveles
		get ERROR()		{ return $ERROR },	// Unexpected Error
		get WARNING()	{ return $WARNING },	// Unexpected Error, Application handles it partaily or completley
		get INFO()		{ return $INFO },	// Expected, but relevant
		get DEBUG()		{ return $DEBUG },	// All stuff of debug
		get file(){
			var file = getLogFolder();
			file.append($profile)
			return file;
		},
		get pathToLogFolder(){
			var result = [];
			for (var i=0,length=$pathToLogFolder.length;i<length;i++)
				result.push($pathToLogFolder[i]);
			return result;
		},
		get path()				{return $file.path},		
		get parentLogger()		{return $parentLogger},
		get registered()		{return $registered},
		get profile()			{return $profile},
		get numOfHistoryLogs()	{return $numOfHistoryLogs},		
		// functions
		error		: function(msg){
			this.log(msg,this.ERROR);
		},
		warning		: function(msg){
			this.log(msg,this.WARNING);
		},
		info		: function(msg){
			this.log(msg,this.INFO);
		},
		debug		: function(msg){
			this.log(msg,this.DEBUG);
		},
		writePrefs	: function(ffPreferences){
			var branch = ffPreferences.getBranch("foxTab.");
			var children = branch.getChildList("", {});
			for (var i=0;i<children.length;i++) {
				if (branch.getPrefType(children[i])==128) {
				    var str = "    "
				    if (branch.prefHasUserValue(children[i])) {
					str = " >> "
				    }
				    this.log(str + children[i] + " = " + branch.getBoolPref(children[i]));
				} else if (branch.getPrefType(children[i])==32) {
				    var str = "    "
				    if (branch.prefHasUserValue(children[i])) {
					str = " >> "
				    }
				    this.log(str + children[i] + " = " + branch.getCharPref(children[i]));
				}
			}
		},		
		/**
		 * Examples:
		 * 1) logger level: ERROR   , message level: WARNING   => not logging
		 * 2) logger level: WARNING , message level: ERROR     => logging
		 * When registered with a parent logger
		 * 3) logger level: WARNING	message level: ERROR     parent level: WARNING => logging both here and in parent
		 * 4) logger level: ERROR   message level: WARNING   parent level: ERROR   => not logging here niether and in parent
		 * 5) logger level: INFO  	message level: WARNING   parent level: ERROR   => logging here but not in parent
		 * In no circumstances a log entry will be logged in parent but not in child
		 */
		log	: function(msg,level){
			if (this.disabled)
				return;
			if (level == null)
				var level=this.level;	// level is default to logger's level
			// log message if it is below threshold level
			if (level > this.level)	
				return;
			if (msg == null)
				var msg="";
			var formatted = (this.showLevel ? LEVEL_TXT[level] + ": " : "");
			if (msg.stack != null && msg.message != null)	// that can be error, or user object with stack and message properties
				//formatted+= "\nmessage: " + msg.message + "\nstack trace: " + msg.stack + "\n";
				formatted+= "\nstack trace\n " + msg.stack + "\n";
			else
				formatted+= msg + "\n";
			$converter.writeString(formatted);
			// propagate msg to parent logger (parent will ignore it, if its level above threshold)
			if ($parentLogger)
				$parentLogger.log(msg,level);
		},
		/**
		 * Register a logger with a parent logger
		 * all logs will be propogated to the parent also
		 * Logger can be registered with null, or with another already registered logger
		 * Logger can not be registered twice, or unregistered
		 * :The two last rules guarantee that there will not be any loops in logger registration (meaning infinite loops)
		 *  if you read it, and believe a loop registration is still possible, contact with the author
		 * @param {Logger} parentLogger parent logger will also log events
		 * @author Eyal eyalm@foxtab.com
		 * @see #registered
		 */
		register : function(parentLogger){
			if (!parentLogger)
				return;
			if (this.registered && $parentLogger.registered)	// this prevents parent being registered with its child
				if (parentLogger === $parentLogger) // no problem, caller simply trying to register the same logger again
					return; // do nothing
				else
					return; //throw already registered
			$registered=true;		
			$parentLogger = parentLogger;
		},
		/**
		 * init a logger
		 * @param feedID the id of the feed (unique and static)
		*/
		init : function(pathToLogFolder,profile,numOfHistoryLogs){
			delete this.init;
			if (!profile || !pathToLogFolder)
				return;
			for (var i=0,length=pathToLogFolder.length;i<length;i++)
				$pathToLogFolder.push(pathToLogFolder[i]);
			$profile=profile;
			
			if (!numOfHistoryLogs || numOfHistoryLogs<1)
				$numOfHistoryLogs=1;
			else
				$numOfHistoryLogs=numOfHistoryLogs;
			pushLog($numOfHistoryLogs-1,true);
			$file = getLogFolder();
			$file.append($profile);
			$file.create(Ci.nsIFile.NORMAL_FILE_TYPE,0777);
			// setup
			var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"].
							 createInstance(Components.interfaces.nsIFileOutputStream);

			// use 0x02 | 0x10 to open file for appending.
			foStream.init($file, 0x02 | 0x10, 0666, 0); 
			// if you are sure there will never ever be any non-ascii text in data you can 
			// also call foStream.writeData directly
			$converter = Cc["@mozilla.org/intl/converter-output-stream;1"].createInstance(Ci.nsIConverterOutputStream);
			$converter.init(foStream, "UTF-8", 0, 0);
			//converter.close(); // this closes foStream			
		}
	}
}